/* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/CodGenC1.c,v
   Comment: Runtime library (including interpreter) for programs gnerated
	    with CodGenC1 code generator.
   Release: scope-0-10 

     Copyright (C) 2001-2002 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   NOTES:

   1. Currently if guards are not used (SCOPE_GUARDS undefined) we only exclude
   some construct to ensure type correctness (i.e. we don't refer to dispatcher
   which is not generated in such case. However, we still keep some code related 
   to guards because it depends to strongly on the way we handle actions 
   (UNMANGLE_ACGD). If this is dependent on using guards then runtime-representation
   could also benefit (if guards are absent). I do not thing you would save even a 
   single byte in trans array (absent guards do not take place anyway) but the 
   interpretation woulod become much simpler.

   CodGenC1.c,v 1.62 2002/11/15 23:40:03 wasowski Exp
*/

#include <assert.h>

static void     enterand( andstref s );
static andstref enteror_tl( orstref s );


#ifndef NDEBUG
unsigned is_andstate( andstref what );
unsigned is_orstate( andstref what );
int conf_member_ext(conf c, andstref s);
#endif

/* Sizes of configurations in use
 **/

/* next_conf size */
static confiter neco_size = 0;
/* prev_conf size */
static confiter prco_size = 0;


#ifndef NDEBUG

/* check if targets are empty (used in assertions) */

int targets_empty (void) 
{
	tgtsiter i = 0;
	for (;i < TGT_WIDTH; ++i) 
		if (targets[i] != STMRK) 
			return 0;
	return 1;
}

/* check if conf is empty (used in assertions) */

int conf_empty (conf c)
{
	if (c == next_conf) return (neco_size == 0); 
	else return (prco_size == 0);
}

/* check if state s is a member of c */

int conf_member_ext(conf c, andstref s)
{
        confiter i=0;
	confiter size = (c == next_conf) ? neco_size : prco_size;

        for (; i < size; ++i) 
		if (c[i] == s) 
			return 1;
        return 0;
}

#if 0
/* unused ? */
/* check if c1 << c2 */
  
int conf_subset(conf c1, conf c2)
{
	confiter i = 0;
	for (; i < STATE_WIDTH; i++)
		if ( c1[i] != STMRK && !conf_member_ext(c2,c1[i]) ) 
			return 0;
	return 1;
}
#endif

void next_conf_reset(void) {

	confiter i = 0;
	neco_size = 0;
}

int compar(const void * a, const void * b)
{
	if ( *(andstref*)a<*(andstref*)b ) 
		return -1;
	else if (*(andstref*)a==*(andstref*)b) 
		return 0;
	return 1;
}

#include <stdlib.h>
#include <stdio.h>

void print_and_sym( andstref s ) 
{
	int c = 0;
	while (1) {
		if (and_syms[c].id == s) {
			printf ("%s ", and_syms[c].name );
			break;
		} else ++c;
	}
}

/* check if what is an andstate. Only works in --debug mode, bcs
 * it uses symbolic dictionaries 
 */

unsigned is_andstate( andstref what ) 
{
	int c = 0; 
	while (and_syms[c].id != 0) {
		if (and_syms[c].id == what) 
			return 1;
		c++;
	}
	return 0;
}

/* check if what is an orstate. Only works in --debug mode,
 * bcs it uses symbolic dictionaries
 */

 unsigned is_orstate( orstref what )
 {
	int c = 0;
	while (or_syms[c].id != 0) {
		if (or_syms[c].id == what)
			return 1;
		c++;
	}
	return 0;
 }

void dump_conf() 
{
	int i = 0;
	conf myconf;
	for (i=0; i < neco_size; ++i) 
		myconf[i] = next_conf[i];

	printf ("conf: [");
	qsort(myconf, neco_size, sizeof(andstref), compar);
	for (i=0; i < neco_size; ++i)
		if (myconf[i] != STMRK) print_and_sym(myconf[i]);
	printf("]\n");	
}


#endif /* NDEBUG */



/* test state activity (what is an and-state identifier).
 * Variable what plays role of lower bound throughout the function.
 * you can substitute it with min if you want
 */

VS_BOOL activeand( const andstref what )
{
	andstref max;
	confiter i = 0;
	const horcell * parent = GET_OR(*GET_AND(what));
	const horcell *j  = parent+1;

	/* you should never ask for sthg which definitely is not a state */
	assert ( what != STMRK );
	assert ( what != HMRK  );
	assert ( what != IMRK  );
	assert ( is_andstate(what) );
	/* you should never need to test activity of rootstate */
	assert ( *GET_AND(what) != STMRK );
	/* in other words (should be a redundant test): */
	assert ( what != ROOT_STATE );
	
	for (; *j != what; ++j);
	
	/* 'what' must be a child of its own parent */
	assert (*j != STMRK);  
	/* what and its next sibing mustn't violiete DFS (preorder) numbering */
	assert (*(j+1) == STMRK || *(j+1) > what );

	if (*(j+1) == STMRK) {
		max = *parent;
		assert ( max >= what );

	} else {
		max = (*(j+1)) -1;
		/* printf ("what %d max %d \n", what, max); */
		assert ( max >= what );
		/* repetition to get more precise diagnosis */
	}


	/* andstate activity is a subset of its parent activity */
	assert ( max <= *GET_OR(*GET_AND(what)) );
	assert ( *GET_AND(what) == STMRK || what >= *(GET_OR(*GET_AND(what))+1) );
	/* non-state will never cause activity */
	assert ( STMRK < what );

	for (; i < prco_size; ++i)
		if (prev_conf[i] >= what && prev_conf[i] <= max) 
			return 1;
	return 0;
}

#ifdef SCOPE_SIGNALQUEUE

/* Signal Queue - implemented as a round buffer */

static queueiter qbeg = 0;
static queueiter qend = 0;

/* silently assume that queue is long enough!
 * which shall be the case, since verification
 * returns this result
 **/

void enqueue1 ( eventtag e ) {

	assert ( e < EVENTNO );

	queue[qend] = e; 

	if (qend == QUEUE_LENGTH - 1) qend = 0;
	else qend++;

	assert( qend != qbeg );
}

#ifdef SCOPE_SLIST
void enqueueList ( signalref ref ) {

	  while (signals[ref]!=NOEVENT)
		  enqueue1(signals[ref++]);
		  
}
#endif

/* returning NOEVENT means that the queue is empty 
 * result is returned in CurrEvent.tag (global variable)
 * That is the way you do it in this funny language to conserve space ...
 **/

void dequeue1 (void) {


	if (qend != qbeg) /* non empty */ {

		CurrEvent.tag = queue[qbeg];
		if (qbeg == QUEUE_LENGTH -1) qbeg = 0;
		else qbeg++;
	} else
		CurrEvent.tag = NOEVENT;
}

#endif /* SCOPE_SIGNALQUEUE */


/* Implementation of The Interpreter 
 **/

/* add s to configuration next_conf */

void conf_add(andstref s) 
{
	assert(is_andstate(s));	
	assert(!conf_member_ext(next_conf,s));
	assert(neco_size < STATE_WIDTH );

	next_conf[neco_size++] = s;
}




/* entering states. Invariant is that if called on the whole hierarchy
 * leaves vector of target empty. If called on the part removes the states
 * belonging to this part. This is deep entering, so it sometimes has
 * to search through the hierarchy to decide if the state should be entered.
 **/

/* to decrease tail recursion problems we return the id of andstate to be entered
 * this should be applied to enter_and in trampoline fashion.
 */
static andstref enteror_tl ( orstref scope )
{
	const horcell * i = (GET_OR(scope)+1);
	andstref min = *i;
	andstref max = *(i-1);
	tgtsiter j = 0;

#ifdef SCOPE_HISTORY
	historyref h = *(GET_OR(scope)-1);
#endif

        for (; j < TGT_WIDTH; ++j) 

		if (targets[j] <= max && targets[j] >= min) {

			for (;*i <= targets[j] && *i !=STMRK; ++i );
			--i;
			if (targets[j] == *i) 
				targets[j] = STMRK;
			exec(GET_ENTER(*i));
			return *i;
		}

	/* enter by default: history or initial */
#ifdef SCOPE_HISTORY
	if ( h != STMRK ) {

		exec (GET_ENTER(history[h]));
		return history[h];

	} else {
#endif 
		exec (GET_ENTER(*(GET_OR(scope)+1)));
		return *(GET_OR(scope)+1);
#ifdef SCOPE_HISTORY
	}
#endif 

}

/* a macro defining enteror using enteror_tl
 * working around tail recursion problems in C
 * Could be avoided if the compiler is known to do it automatically
 * FIXME: for instance an option to code generator (use or no use for 
 * tail recursion).
 */

#define enteror(s) enterand(enteror_tl(s))

/* unfortunately tail recursion cannot be nicely used in
 * entering and-states
 */

static void enterand ( andstref scope )
{
	const handcell * i = GET_AND(scope)+1;

	if ( *i == STMRK ) conf_add (scope);
	else {
		do enteror( *(i++) );
		while (*i != STMRK);
	}
}

#ifndef IMPURE_EXIT 

#define exitor_impure(s)	exitor_pure(s)

#else /* defined (IMPURE_EXIT) */

static void exitand_impure( andstref s );

/* remove s from configuration */

void conf_del( andstref s )
{
	confiter i = 0;
	
	assert(is_andstate(s));
	assert(conf_member_ext(next_conf,s));
	assert(!conf_empty(next_conf)); 

	while (next_conf[i] != s) {
		++i;
		assert (i < STATE_WIDTH);
		assert (i < neco_size);
	}
	next_conf[i]          = next_conf[--neco_size];
}

/* exiting state (recursively exit it substates)
 **/




/* trampoline cannot be used also in this case ->
 * exit actions must be called after recursive call.
 */
static void exitor_impure( orstref scope )
{
	confiter j = 0;
	const horcell * i = GET_OR(scope) + 1; /* get first child */
	andstref min = *i;
	andstref max = *(i-1);

#ifdef SCOPE_HISTORY
	historyref h = (historyref)*(i-2);
#endif

	assert (*i != STMRK);		 /* orstate cannot be basic */


	for (; j < neco_size; ++j)
		if (next_conf[j] <= max && next_conf[j] >= min)  {
			for(; *i <= next_conf[j] && *i != STMRK; ++i);
			--i;
			exitand_impure (*i);
			exec( GET_EXIT(*i) );
#ifdef SCOPE_HISTORY
			if (h != STMRK)
				history[h] = *i;
#endif
			return;
		}
	
}


static void exitand_impure( andstref scope )
{ 
	const orstref * i = GET_AND(scope)+1;
        if ( *i == STMRK ) { /* BASIC */
		conf_del( scope );
        } else {
		do exitor_impure( *i );
		while (*(++i) != STMRK);
	}
}

#endif  /* IMPURE_EXIT */

#ifndef PURE_EXIT

#define exitor_pure(s) exitor_impure(s)

#else  /* defined(PURE_EXIT) */

void  exitor_pure( orstref s ) 
{
	const andstref * i = GET_OR(s);
	andstref min = *(i+1);
	andstref max = *i;
	confiter j = 0;

	#ifndef NDEBUG
	int removed = 0;
	int presize = neco_size;
	#endif

	assert(is_orstate (s));
	assert(!conf_empty(next_conf));

	if (min == STMRK) 
		min = max;

	while (j < neco_size) 
		if (next_conf[j] >= min && next_conf[j] <= max) {
			next_conf[j] = next_conf[--neco_size];
			#ifndef NDEBUG
			removed++;
			#endif
		} else j++;

	assert (removed);
	assert (neco_size + removed == presize);
}

#endif /* PURE_EXIT */

/* fire a transition. Transition is assumed to be enabled.  The parameter
 * passed is assumed to be the index in transitions array were the
 * guard/action indicators are kept. We point into the middle of the record
 * for slight efficiency gain. The preceding part will not be used in this
 * function anyway.
 *
 * The assumption is that targets are empty on entry and on exit.
 **/ 

void fire (const transcell * ptr) 
{
	actionref ac = STMRK;
	const transcell * s;
	

#ifdef SCOPE_GUARDS
	guardref  gd = STMRK;
	UNMANGLE_ACGD(ptr,ac,gd);
	if ( !eval(gd) ) return;
#else
	UNMANGLE_AC(ptr,ac);
#endif 
	assert (targets_empty());

	s = ptr;

	while (*s != STMRK) {

		assert(*s < MARKNO);

#ifndef IMPURE_EXIT
		assert (*s != IMRK && *s != HMRK );
#endif
#ifndef PURE_EXIT
		assert (*s != FPSCOP && *s != NFPSCOP );
#endif

		switch (*(s++)) {
#ifdef IMPURE_EXIT
			case HMRK:  /* flat non pure exit */

				do exitor_impure (*(GET_AND(*(s++))));
				while (*s >= MARKNO);
				break;

			case IMRK: /* nonflat and nonpure exit */

				exitor_impure (*(s++)); 
				while (*s >= MARKNO) s++;
				break;
#endif
#ifdef PURE_EXIT
			case FPSCOP: /* flat pure exit */

				do exitor_pure (*(GET_AND(*(s++))));
				while (*s >= MARKNO);
				break;

			case NFPSCOP: /* nonflat and pure exit */

				exitor_pure (*(s++));
				while (*s >= MARKNO) s++;
#endif 
		}

		assert(*s < MARKNO);
	}

	exec(ac);

	while (*ptr != STMRK) {
	 
		assert( *ptr < MARKNO );

#ifndef IMPURE_EXIT
		assert (*s != IMRK && *s != HMRK );
#endif

#ifndef PURE_EXIT
		assert (*s != FPSCOP && *s != NFPSCOP );
#endif

		if (
#ifdef IMPURE_EXIT		
				*ptr == HMRK ||
#endif

#ifdef PURE_EXIT
				*ptr == FPSCOP 
#else
				0
#endif
		   ) { /* flat */
	
			while (*(++ptr) >= MARKNO) {
				targets[0] = *ptr;
				enteror(*(GET_AND(*ptr)));
			}

		} else { /* non flat */

			tgtsiter c = 0;
			orstref scope = *(++ptr);
			for (++ptr; *ptr >= MARKNO; ++ptr, ++c)
				targets[c] = *ptr;
			enteror(scope); 
		}

		assert (targets_empty());
		assert (*ptr < MARKNO);
	}
}

/* microstep - assumes that proper event is already in CurrEvent
 * no conflict resolution currently. Everything will collapse on conflict
 */

void microstep (void) {

      const transcell * t = trans + tranidx[CurrEvent.tag];

      assert( CurrEvent.tag < NOEVENT );

      if (tranidx[CurrEvent.tag] != TRANS_MAX) do { 

	      unsigned int pc,nc;
	      UNMANGLE_PCNC(t,pc,nc);
      
	      for(; pc > 0; --pc, ++t ) 
		      if (!activeand(*t)) goto skiptr;

	      for(; nc > 0; --nc, ++t )
		      if (activeand(*t)) goto skiptr;

	      fire(t); /* checking guard is deferred to fire */
	      
      skiptr: while (*t != STMRK) ++t;
	      ++t;
		
      } while (t < trans + tranidx[CurrEvent.tag+1]);
}


/* macrostep - assumes that external event is already in CurrEvent 
 * the newest configuration is always kept in next_conf. prev_conf
 * is only occasionally used for guard computations.
 **/
void macrostep (void) {

#ifdef SCOPE_SIGNALQUEUE
        while (CurrEvent.tag != NOEVENT) {
#endif 
          confiter i = 0; 
	  prco_size = neco_size;
	  for(; i < neco_size; ++i) 
		  prev_conf[i] = next_conf[i]; 

	  assert (!conf_empty(next_conf));

	  microstep();

#ifdef SCOPE_SIGNALQUEUE
	  dequeue1();
        } 
#endif

}
	
/* let us initialise the model 
 * Note: st_init will only work on empty configuration.
 * If there is a need for multiple initializations at runtime
 * another function (see testdrv) has to be called to clean current conf
 **/

void st_init(void)
{
        enterand (ROOT_STATE);

#ifdef SCOPE_SIGNALQUEUE
        dequeue1();
        macrostep();
#endif
}


