(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/CodGenC1_front.sml,v
   Comment: Front end of CodGenC1 code generator. Takes an Absyn model and
	    compiles it to lrsystem (intermediate representation)
   Release: scope-0-10 

     Copyright (C) 2001-2002 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   NOTES:

   This modules tries to avoid operating on numbers (addresses, references,
   etc). Symbolic names are used as far as possible and actual array
   addresses are assigned in scheduler. One exception is actions and guards,
   as we need to count the number of unique before type inference. This is
   conveniently done by constructing an ulist, which performs a scheduling in
   place.

   CodGenC1_front.sml,v 1.31 2002/11/09 21:57:43 wasowski Exp
*)


(* some interface *)

open Haskellish List Say Types CodGenC1_LR Absyn
infixr 1 $; 
infix 3 --;

val alloc = UList.alloc

type actions = CodGenC1_LR.actions


(* Types *)

(* type of textual substitutions *)

type subst = symbol env

(* implementation *)

(* substitution application *)

fun subst_exp (subst :subst) (e :exp) :exp = 
	case e of Access a => Access (subst_acc subst a)
		| Prim1(ope,e) => Prim1(ope,subst_exp subst e)
		| Prim2(ope,e1,e2) => 
			Prim2(ope, subst_exp subst e1, subst_exp subst e2)
		| Call oe => Call (subst_oexp subst oe)
		| anything => anything

and subst_acc (subst :subst) (a :access) :access =
	case a of AccessVar v => 
			AccessVar (lookup subst v handle Subscript=> v)
		| AccessArr (a,e) =>
			AccessArr (lookup subst a handle Subscript=> a,
				   subst_exp subst e)

and subst_oexp (subst :subst) (f,ps) :oexp = (f,map (subst_exp subst) ps)
	
fun subst_assgn (subst :subst) (a,e) :assgn = 
	(subst_acc subst a, subst_exp subst e)

fun subst_sexp (subst :subst) (e,ps) :sexp = (e, map (subst_exp subst) ps)

fun subst_action (subst :subst) {assgns, outputs, signals} :action =
	{ assgns = map (subst_assgn subst) assgns,
	  outputs = map (subst_oexp subst) outputs,
	  signals = map (subst_sexp subst) signals }

fun subst_guard (subst :subst) guard :guard =
	map (subst_exp subst) guard

(* compute substitution for transition *)

fun comp_subst (ets :evtype env) (e,ps) :symbol env =
	(case lookup ets e of 
	  (true,_) => 
		empty before
		(if null ps then () else warning ("Parameters for signal "
						^ Symbol.name e ^ "ignored"))
	| (false,[]) => empty
	| (false,ts) =>
		let val uf = "CurrEvent.fields._E_" ^ Symbol.name e ^ ".f"
		    val fields =tabulate (length ts,
			fn i => Symbol.symbol(uf^ Int.toString i))
		in bindZip empty (ps,fields)  end
	)handle Subscript => impossible "#20020709c"

(* adressing mode selector for targets with scopes *)

fun isflat ([s1],s2) :bool = parentand s1 = nameor s2
  | isflat _ = false  

(* FIXME: this function is whole lot too
   ugly. It calls for rethinking and reimplementation. *)

fun comp_targets []      prev (res,mode) :lratom list * exitmode = (res,mode)
  | comp_targets (t::ta) prev (res,mode) =	(* YUCK: this is slow and really    *)
						(* calls for syntax tree annotations*) 
	case (isflat t,Stat.is_expureor (#2 t)) of 

		  (true,false) => 
			if prev = HMRK then comp_targets ta HMRK ((LRAndStateRef $ nameand $ hd $ #1 t)::res, use_impure mode)
			else comp_targets ta HMRK ((LRAndStateRef $ nameand $ hd $ #1 t)::LRMrk HMRK ::res, use_impure mode)
		| (false,false) =>
			comp_targets ta IMRK
			(foldl (fn (s,e) => LRAndStateRef(nameand s)::e) 
			((LROrStateRef $ nameor $ #2 $ t) :: LRMrk IMRK::res) (#1 t), use_impure mode)
		| (true,true) =>
			if prev = FPSCOP then comp_targets ta FPSCOP ((LRAndStateRef $ nameand $ hd $ #1 t)::res,use_pure mode)
			else comp_targets ta FPSCOP ((LRAndStateRef $ nameand $ hd $ #1 t) :: LRMrk FPSCOP ::res,use_pure mode)
		| (false,true) =>
			comp_targets ta NFPSCOP 
			(foldl (fn (s,e) => LRAndStateRef(nameand s)::e)
			((LROrStateRef $ nameor $ #2 $ t) :: LRMrk NFPSCOP ::res) (#1 t),use_pure mode)
			
(* compile a single transition, suitable for folding *)

fun comp_tran (ets :evtype env) (t :stran, (tenv :tenv, guards, actions, mode)) 
	: tenv * guards * actions * exitmode =
	let val subst = comp_subst ets (#trigger t) 
	    val (actref,actions1) = 
			alloc (subst_action subst (#action t),actions)
	    val guard = if null (#guard t) then always else (#guard t)
	    val (guardref,guards1) =
			alloc (subst_guard subst guard, guards)
	    val (targets,mode1) = comp_targets (#targets t) NFPSCOP ([],mode)
	    val trans  = lookup tenv (#1 $ #trigger t) 
				handle Subscript => [] 
	    val trans1 = List.revAppend (targets, LRMrk STMRK::trans)
	    val trans2 = LRAcGd (actref+markno,guardref+markno) ::trans1
	    val trans3 = map (LRAndStateRef o nameand) (#pos t @ #neg t) @trans2
	    val trans4 = LRCondCounters(length (#pos t), length (#neg t))::trans3
	    val tenv1  = bind1 tenv (#1 $ #trigger t, trans4)
	in (tenv1, guards1, actions1,mode1) end
	handle Subscript => impossible "#20020709b"

(* compile a transitions list (already degrouped, no dynmic scopes) *)

fun comp_trans (ets: evtype env) (actions :actions) (trans :stran list) 
	:tenv * guards * actions * exitmode =
	let val guards = #2 (alloc (always,UList.initialize guardsize))
	    val nomode = { pure = false, impure = false }
	in foldl (comp_tran ets) (empty,guards,actions,nomode) trans end
	handle Subscript => impossible "#20020709a"

(* compiling states. functions should be used with hfolds *)

fun compor (ORState s, (hor,hand,actions,history))
	:lratom list * lratom list * actions * lratom list =
	let val kids    = map nameand (#children s)
	    val hor1    = map LRAndStateRef kids @ LRMrk STMRK ::hor
	    (* FIXME: yuck! RMDor call is extremely inefficient as the
		      information is never reused up in the heirarchy *)
	    val lastkid = nameand (RMDor (ORState s))
	    val hor2    = LRStateLabel(#name s) ::LRAndStateRef lastkid::hor1
			  handle Empty => intErr Empty 200210290
	    val (hor3,history1) = 
			if #marking s <> History then (hor2,history)
			else (LRHistoryRef (length history) ::hor2,
			      LRAndStateRef (hd kids) ::history)
	in (hor3, hand, actions,history1) end
		
fun compand (ANDState s,(hor,hand,actions,history))
	:lratom list * lratom list * actions * lratom list = 
	let val (enref,actions1) = alloc (#entry s,actions) 
	    val (exref,actions2) = alloc (#exit s,actions1)
	    (* +(markno) below reserves space for markers *)
	    val hand1 = LRMrk STMRK ::hand
	    val hand2 = foldr (fn (s,e) => LROrStateRef (nameor s)::e) 
				hand1 (#children s)
	    val hand3 = LROrStateRef (valOf $ #parent s) ::hand2
			handle Option => LRMrk STMRK ::hand2
	    val hand4 = LREnEx (enref+markno,exref+markno) 
				::LRStateLabel (#name s) ::hand3
	in (hor,hand4,actions2,history) end 

(* compile signals. Only look at nonempty, nonsingle signals.
   The rest is optimized anyway *)

fun comp_signals (actions :actions) :signals =
	let val actionsl = UList.toList actions 
	    fun comp_signal ({signals=[],...},e) = e
	      | comp_signal ({signals=[s],...},e) = e
	      | comp_signal ({signals,...},e) =
		let val sigs = map (LREventRef o #1) signals
		in #2 $ alloc (LRNoEvent::sigs,e) end
	in foldl comp_signal (UList.initialize sigsize) actionsl end


(* compile hierarchy. result is: hor and hand arrays in lr format. list of
   actions in internal representation, history state vector in lratom format *)

fun comp_states (h :hierarchy) 
	:lratom list * lratom list * actions * lratom list =
	let val actions = #2 (alloc (null_action,UList.initialize actionsize))
	    val history = [LRMrk STMRK]
	    val e       = ([],[], actions, history)
	    val (hor,hand,actions,history) = 
			hfoldand compor compand e (gettree h)
	    val hand1   = List.tabulate(markno, fn _ => LRMrk STMRK) @ hand
	in  (LRMrk STMRK::hor, hand1, actions,history) end

(* propose length for signal queue - 0 means no queue *)

fun comp_queuelen (queuelen :int) (evtypes :evtype env) :int option =
	if List.exists (#1 o #2) (SE.toList evtypes) 
	then SOME queuelen 
	else NONE
	
(* compile system, compute some sizes and infer types *)

(* FIXME: static scope and normalise should be called way *before* the code
   generator. *)

fun comp_system (queuelen :int) (sys :system) :lrsystem =
	let val (_,h) = Normalise.pull_defs (#denv sys, #hierarchy sys)
	    val (hor,hand,actions1,history) = comp_states h
	    val t1 = Normalise.degroup_all (#evgroups sys) (#trans sys)
	    val t2 = map (Normalise.rmRedundantTgts h) t1
	    val sts = StaticScope.static_scopes_all h t2
	    val (tenv,guards,actions2,exitmode)= 
			comp_trans (#evtypes sys) actions1 sts
		(* safe to call on non transformed system *)
	    val (orsts,andsts) = Stat.state_dist sys
	    val _ = assert (fn ()=> orsts < andsts) "|State_or| >= |State_and|"
	in { name     = Symbol.name (#name sys),
	     hor      = hor,
	     hand     = hand,
	     root     = nameand (gettree h),
	     tenv     = tenv,
	     
	     history  = history,
	     signals  = comp_signals actions2,
	     actions  = actions2,
	     guards   = guards,

	     evtypes  = #evtypes sys,
	     vartypes = #vartypes sys,
	     venv     = #venv sys,
	     cenv     = #cenv sys,

	     queuelen = comp_queuelen queuelen (#evtypes sys),
		(* safe to call on non transformed system *)
	     statew   = Stat.conf_card_bound sys, 
	     targetw  = Stat.targets_card_bound sts,
	     orsts    = orsts,
	     andsts   = andsts,
	     exitmode = exitmode }
	end

(* infer type sizes for given system in LR *)

fun resolve_types (sys :lrsystem) :lrsizes = 
	let val evnumber  = SymbolEnv.count (#evtypes sys)
	    val signumber = UList.length (#signals sys)
	in { 
		actionref = Types.mintysize (UList.length(#actions sys) + markno -1),
		guardref  = Types.mintysize (UList.length(#guards  sys) + markno -1),
		signalref = Types.mintysize (evnumber + 1 + signumber - 1),
		(* signalrefsize can be done better, substracting 
		   the length of the last entry *)
		orstref   = mintysize (#orsts sys + 1 - 1),
		andstref  = mintysize (#andsts sys+ markno - 1),
		evnumber  = evnumber }
	end



