(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/Types.sml,v
   Comment: Modelling types which are shared along vsr.sml and Absyn.sml 
   Release: scope-0-10

     Copyright (C) 2001-2002 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   WARNING: Prior to v 1.2 the meaning of this file was different (these
   were types as used  in hvsr.sml). Only the name has been kept.

   Types.sml,v 1.10 2002/11/04 21:13:55 wasowski Exp
*)

structure Types :> Types =
struct

  open Err Haskellish
  infixr 0 $

  datatype simplety = 
		VoidT  | VoidPtrT | BoolT  | CharT   | UCharT |
		IntT   | Int8T    | Int16T | Int32T  | UIntT  | 
		UInt8T | UInt16T  | FloatT | DoubleT   

  datatype ty = 
	  SimpleType of simplety	
        | ArrayType  of simplety * int	(* cell type, size  *)
  
  (* variable: internal(static), type *)

  type vartype = bool * ty                 

  (* event: internal, params  *)
  
  type evtype = bool * simplety list	   
  
  (* function prototype *)

  type otype = simplety * simplety list  (* output: return, params   *)
  datatype value = ValI of int    (* for all integers, pointers and chars *)
		 | ValR of real 
		 | ValA of value list

  (* defualt types for values *)	

  fun defvalty VoidPtrT = ValI 0
    | defvalty BoolT    = ValI 0
    | defvalty CharT	= ValI 0
    | defvalty UCharT   = ValI 0
    | defvalty IntT     = ValI 0 
    | defvalty Int8T    = ValI 0
    | defvalty Int16T   = ValI 0
    | defvalty Int32T   = ValI 0
    | defvalty UIntT    = ValI 0 
    | defvalty UInt8T   = ValI 0
    | defvalty UInt16T  = ValI 0
    | defvalty FloatT   = ValR 0.0
    | defvalty DoubleT  = ValR 0.0
    | defvalty VoidT	= 
	impossible "Default value should never be queried for type void."

 fun tyname VoidT    = "VS_VOID"    
   | tyname VoidPtrT = "VS_VOIDPTR"
   | tyname BoolT    = "VS_BOOL"   
   | tyname CharT    = "VS_CHAR"   
   | tyname UCharT   = "VS_UCHAR"  
   | tyname IntT     = "VS_INT"    
   | tyname Int8T    = "VS_INT8"   
   | tyname Int16T   = "VS_INT16"  
   | tyname Int32T   = "VS_INT32"  
   | tyname UIntT    = "VS_UINT"   
   | tyname UInt8T   = "VS_UINT8"  
   | tyname UInt16T  = "VS_UINT16" 
   | tyname FloatT   = "VS_FLOAT"  
   | tyname DoubleT  = "VS_DOUBLE" 



  (* Simplified C type coersions *)

  fun int2real (i :int) :value = ValR $ Real.fromInt i

  fun real2int (r :real) :value = ValI $ Real.trunc r

  fun bool2int false = ValI 0
    | bool2int true  = ValI 1
 
  fun real2bool (r :real) :bool = 
	if r < 0.0 then floor r <> 0 else ceil r <> 0

 
  (* behaviour of floats in boolean conditions is slightly peculiar.
     They do not follow standard cast policy (truncating) in this place *)

  fun val2bool (ValI i) = i <> 0
    | val2bool (ValR r) = real2bool r
    | val2bool (ValA a) = 
	runtimeErr "Array addresses may not be cast to bools in VisualSTATE."



  (* promote types so they match - always lossless *)

  fun promote (ValA _,_) = 
	runtimeErr "Array addresses are illegal as scalars in visualSTATE"
    | promote (_,ValA _) =
	runtimeErr "Array addresses are illegal as scalars in visualSTATE"
    | promote (ValR a,ValR b) = (ValR a,ValR b)
    | promote (ValI a,ValI b) = (ValI a,ValI b)
    | promote (ValI a,ValR b) = (int2real a, ValR b)
    | promote (ValR a,ValI b) = (ValR a, int2real b)



  (* get integer out of value with *no* cast *)

  fun getInt (ValI i) = i
    | getInt (ValR i) = 
	runtimeErr "Integer value expected where real found!"
    | getInt (ValA i) = 
	runtimeErr "Integer value expected where non-scalar is found"

  (* get the minimum size of integer type (in bytes) to represent n *)

  fun mintysize (i :int) :int = 
	if i <= 255 then 1
	else if i <= 255*255 then 2
	else if i <= 255*255*255 then 4
	else impossible "Values more than 64 bits are not handled currently"
  
  (* returns unsigned integer type of given size (in bytes) *)

  fun uint_bysize 1 = UInt8T
    | uint_bysize 2 = UInt16T
    | uint_bysize _ = 
	impossible "#20020213a: request for integer type of incorrect size"
 
end

